/* 
 * Copyright (C) 2000 the xine project
 * 
 * This file is part of xine, a unix video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: main.c,v 1.32 2000/12/12 00:45:14 guenter Exp $
 *
 * xine main
 *
 */

/* Requiered for getsubopt() */
#define _XOPEN_SOURCE 500

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include "demux.h"
#include "xine.h"
#include "monitor.h"
#include "configfile.h"
#include "video_out/video_out.h"
#include "audio_out/audio_out.h"
#include "gui/gui_main.h"
#include "gui/gui_dnd.h"
#ifdef HAVE_GETOPT_LONG
#  include <getopt.h>
#else
#  include "getopt.h"
#endif

uint32_t xine_debug;

extern Display        *gDisplay;
extern vo_driver_t    *gVideoDriver;
extern Window          gVideoWin;

/* playlist */
int                    nPlayNum;
int                    ncur;
char                  *playlist[100];

/* options args */
static const char *short_options = "?h"
#ifdef HAVE_XV
 "s"
#endif
#ifdef DEBUG
 "d:"
#endif
 "a:A:p::";
static struct option long_options[] = {
  {"help"           , no_argument      , 0, 'h' },
#ifdef HAVE_XV
  {"shared-memory"  , no_argument      , 0, 's' },
#endif
  {"audio-channel"  , required_argument, 0, 'a' },
  {"audio-driver"   , required_argument, 0, 'A' },
  {"auto-play"      , optional_argument, 0, 'p' },
#ifdef debug
  {"debug"          , required_argument, 0, 'd' },
#endif
  {0                , no_argument      , 0,  0  }
};

extern const char* ao_available[];
int ao_use_driver = AO_UNSET;
int curAudioTrack = 0;

/* ------------------------------------------------------------------ */
/*
 * Clean a string (remove spaces and '=' at the begin,
 * and '\n', '\r' and spaces at the end.
 */
char *atoa(char *str) {
  char *pbuf;

  pbuf = str;

  while(*pbuf != '\0') pbuf++;

  while(pbuf > str) {
    if(*pbuf == '\r' || *pbuf == '\n' || *pbuf == '"') pbuf = '\0';
    pbuf--;
  }

  while(*pbuf == '=' || *pbuf == '"') pbuf++;

  return pbuf;
}
/* ------------------------------------------------------------------------- */
/*
 *
 */
void show_banner(void) {

  printf("This is %s - a mpeg 1, 2 player v%s - (c) 2000 by G. Bartsch.\n", 
	 PACKAGE, VERSION);
}
/* ------------------------------------------------------------------------- */
/*
 *
 */
void show_usage (void) {

  printf("\n");
  printf("Usage: %s [OPTIONS]... [MRL]\n", PACKAGE);
  printf("\n");
  printf("OPTIONS are:\n");
#ifdef HAVE_XV
  printf("  -s, --shared-memory          Use X shared memory extension "
	 "for output.\n");
#endif
  printf("  -a, --audio-channel <#>      Select audio channel '#'.\n");
  printf("  -A, --audio-driver <drv>     Select audio driver (oss"
#ifdef HAVE_ALSA
	 ", alsa"
#endif
	 " or null).\n");
  printf("  -p, --auto-play [opt]        Play on start. Can be followed by:\n");
  printf("                                 'f': start in fullscreen mode,\n");
  printf("                                 'h': hide control panel,\n");
  printf("                                 'q': quit when play is done.\n");
#ifdef DEBUG
  printf("  -d, --debug <flags>          Debug mode for <flags> ('help' for list).\n");
#endif
  printf("\n");
  printf("valid MRLs (media resource locator) are:\n");
  printf("  File:  'path/foo.vob'\n");
  printf("         '/path/foo.vob'\n");
  printf("         'file://path/foo.vob'\n");
  printf("         'fifo://[[mpeg1|mpeg2]:/]path/foo'\n");
  printf("         'stdin://[mpeg1|mpeg2]' or '-' (mpeg2 only)\n");
  printf("  DVD:   'dvd://VTS_01_2.VOB'\n");
  printf("  VCD:   'vcd://<track number>'\n");
  printf("\n");
}
/* ------------------------------------------------------------------------- */
/*
 *
 */
#ifdef DEBUG
int handle_debug_subopt(char *sopt) {
  int subopt;
  char *str = sopt;
  char *val = NULL;
  const char *debuglvl[] = {
    "verbose", "metronom", "audio", "demux", 
    "input", "video", "pts", "mpeg", "avi", 
    "ac3", "loop", "gui",
    NULL
  };
  
  while((subopt = getsubopt(&str, debuglvl, &val)) != -1) {
    
    switch(subopt) {
    default:
      /* If debug feature is already enabled, with 
       * XINE_DEBUG envvar, turn it off 
       */
      if((xine_debug & 0x8000>>(subopt + 1)))
	xine_debug &= ~0x8000>>(subopt + 1);
      else
	xine_debug |= 0x8000>>(subopt + 1);
      break;
    }
  }

  if(val) {
    int i = 0;
    fprintf(stderr, "Valid debug flags are:\n\t");
    while(debuglvl[i] != NULL) {
      fprintf(stderr,"%s, ", debuglvl[i]);
      i++;
    }
    fprintf(stderr, "\b\b.\n");
    return 0;
  }

  return 1;
}
#endif
/* ------------------------------------------------------------------------- */
/*
 * Handle audio output multi-args, return first audio plugin load successed.
 * Return AO_UNSET on fail. 
 */

#define FORCENULL  {                                                         \
   fprintf(stderr, "audio driver '%s' failed, using null driver instead.\n", \
           ao_available[aod]);                                               \
   return AO_NULL;                                                           \
		   }
int handle_audioout_subopt(char *sopt) {
  int subopt;
  char *str = sopt;
  char *val = NULL;
  ao_functions_t *aout;
  int aod = 0;
 
  while((subopt = getsubopt(&str, ao_available, &val)) != -1) {
    
    aod = subopt;

    switch(subopt) {
    case AO_OSS:
      if((aout = audio_ossout_init()) != NULL) {
      	return AO_OSS;
      }
      else
	FORCENULL;
      break;

    case AO_NULL:
      if((aout = audio_nullout_init()) != NULL) {
      	return AO_NULL;
      }
      else
	FORCENULL;
      break;

#ifdef HAVE_ALSA      
    case AO_ALSA:
      if((aout = audio_alsaout_init()) != NULL) {
	return AO_ALSA;
      }
      else
	FORCENULL;
      break;
#endif

    }
  }
  
  return AO_UNSET;
}
/* ------------------------------------------------------------------------- */
/*
 *
 */
int main(int argc, char *argv[]) {
  int   option_index = 0;
  int   no_xv;
  char  c = '?';
  char *filename = NULL;
  char *display_name = ":0.0";

#ifdef HAVE_XV
  no_xv = 0;
#else
  no_xv = 1;
#endif

  show_banner();

  xine_debug = 0;
#ifdef DEBUG
  /* If XINE_DEBUG envvar is set, parse it */
  if(getenv("XINE_DEBUG") != NULL) {
    if(!(handle_debug_subopt(atoa((getenv("XINE_DEBUG"))))))
      exit(1);
  }
  
  /*
   * profiling
   */

  profiler_init();
  profiler_set_label (0, "xine_loop     ");
  profiler_set_label (1, "ac3 decoding  ");
  profiler_set_label (2, "mpeg2 decoding");
  profiler_set_label (3, "mp3 decoding  ");
  profiler_set_label (4, "video output  ");
  profiler_set_label (5, "draw frame    ");
  
#endif

  /*
   * parse command line
   */
  opterr = 0;
  while((c = getopt_long(argc, argv, short_options, 
			 long_options, &option_index)) != EOF) {
    switch(c) {

#ifdef HAVE_XV
    case 's': /* Use Xshm insteat of Xv (if Xv is present) */
      no_xv = 1;
      break;
#endif
      
    case 'a': /* Select audio channel */
      sscanf(optarg, "%i", &curAudioTrack);
      /* track_code = track_table[audio_track]; */
      break;

    case 'A': /* Select audio driver */
      if(optarg != NULL) {
	if((ao_use_driver = 
	    handle_audioout_subopt(atoa(optarg))) == AO_UNSET) {
	  int i;
	  fprintf(stderr, "Error: available audio output drivers are:");
	  for(i = AO_OSS; ao_available[i] != NULL; i++)
	    fprintf(stderr, " %s,", ao_available[i]);
	  fprintf(stderr, "\b.\n");
	  exit(1);
	}
      }
      break;

    case 'p':/* Play [[in fullscreen][then quit]] on start */
      xine_user_pref |= PLAY_ON_START;
      if(optarg != NULL) {
	if(strrchr(optarg, 'f')) {
	  xine_user_pref |= FULL_ON_START;
	}
	if(strrchr(optarg, 'h')) {
	  xine_user_pref |= HIDEGUI_ON_START;
	}
	if(strrchr(optarg, 'q')) {
	  xine_user_pref |= QUIT_ON_STOP;
	}
      }
      break;

#ifdef DEBUG      
    case 'd': /* Select debug levels */
      if(optarg != NULL) {
	if(!(handle_debug_subopt(atoa(optarg))))
	  exit(1);
      }
      break;
#endif
      
    case 'h': /* Display usage */
    case '?':
      show_usage();
      exit(1);
      break;
      
    default:
      show_usage();
      fprintf (stderr, "invalid argument %d => exit\n",c);
      exit(1);
    }
  }

  /* If not audio output driver selected, force to OSS */
  if(ao_use_driver == AO_UNSET)
    ao_use_driver = AO_OSS;
  fprintf(stderr, "Using %s audio output driver.\n", 
	  ao_available[ao_use_driver]);

  /* Extract filename (if specified) */
  if (optind < argc) 
    filename = argv[optind];

  config_file_init ();

  /*
   * init X11
   */

  if (!XInitThreads ()) {
    printf ("\nXInitThreads failed - looks like you don't have a thread-save xlib.\n");
    exit (1);
  }

  if(getenv("DISPLAY"))
    display_name = getenv("DISPLAY");

  gDisplay = XOpenDisplay(display_name);

  if (gDisplay == NULL) {
    fprintf(stderr,"Can not open display\n");
    exit(1);
  }

  /*
   * drag and drop
   */

  gui_init_dnd();
  gui_dnd_set_callback (gui_dndcallback);

  /*
   * xine init
   */

  xine_init (no_xv, gui_status_callback);
  xine_select_audio_channel (curAudioTrack);

  /*
   * hand control over to gui
   */

  gui_start(argc-optind, &argv[optind]);

  return 0;
}		

