/* 
 * Copyright (C) 2000 the xine project
 * 
 * This file is part of xine, a unix video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: video_out.c,v 1.22 2000/12/12 21:01:13 guenter Exp $
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "video_out.h"
#include "metronom.h"
#include "monitor.h"
#include "../libmpeg2/mpeg2.h"
#include "../audio_out/audio_out.h"

#define NUM_BUFFERS     20

extern uint32_t xine_debug;

extern ao_functions_t *gAudioOut;

typedef struct img_buf_fifo_s {
  vo_image_buffer_t *first;
  vo_image_buffer_t *last;
  int                num_buffers;
} img_buf_fifo_t;

typedef struct vo_priv_s {

  img_buf_fifo_t    *free_img_buf_queue;
  img_buf_fifo_t    *display_img_buf_queue;

  vo_image_buffer_t *displayed_image; /* image currently displayed */

  int                video_step;
  int                pts_per_half_frame;

  /* statistics */

  int                num_frames_delivered;
  int                num_frames_skipped;
  int                num_frames_discarded;

  int                width, height;
  int                bInitialized;
} vo_priv_t;

vo_driver_t *gVideoDriver;
vo_priv_t    gVO;

img_buf_fifo_t *new_img_buf_queue () {
  img_buf_fifo_t *queue;

  queue = (img_buf_fifo_t *) malloc (sizeof (img_buf_fifo_t));
  queue->first       = NULL;
  queue->last        = NULL;
  queue->num_buffers = 0;

  return queue;
}

void append_to_img_buf_queue (img_buf_fifo_t *queue, vo_image_buffer_t *img) {

  img->next = NULL;

  if (!queue->first) {
    queue->first = img;
  } else {
    queue->last->next = img;
  }
  queue->last  = img;

  queue->num_buffers ++;
}

vo_image_buffer_t *remove_from_img_buf_queue (img_buf_fifo_t *queue) {

  vo_image_buffer_t *img;

  img = NULL;

  if (queue->first) {
    img = queue->first;

    queue->first = queue->first->next;
    if (!queue->first)
      queue->last = NULL;

    queue->num_buffers--;
    
  } else {
    /* printf ("video_out : warning - queue empty\n"); */
  } 
  
  return img;
}

/*
void print_queue (img_buf_fifo_t *queue) {
  
  printf ("first : %d last : %d num : %d\n",
	  queue->first, queue->last, queue->num_buffers);

}
*/

void reallocate_img_buffers () {

  int i;
  vo_image_buffer_t *img;
  
  /*
   * free all images
   */

  if (gVO.displayed_image) {
    gVideoDriver->dispose_image_buffer (gVO.displayed_image);

    gVO.displayed_image = NULL;
  }

  while ((img = remove_from_img_buf_queue (gVO.display_img_buf_queue))) {
    gVideoDriver->dispose_image_buffer (img);
  }
  while ((img = remove_from_img_buf_queue (gVO.free_img_buf_queue))) {
    gVideoDriver->dispose_image_buffer (img);
  }

  /*
  print_queue (gVO.display_img_buf_queue);
  print_queue (gVO.free_img_buf_queue);
  */
  /*
   * allocate new images
   */

  for (i=0; i<NUM_BUFFERS; i++) {
    img = gVideoDriver->alloc_image_buffer ();
    append_to_img_buf_queue (gVO.free_img_buf_queue, img);
  }

}

void vo_set_size (uint32_t width, uint32_t height, uint32_t ratio, int rate_code) {

  if (gVideoDriver->set_size (width, height, ratio)) 
    reallocate_img_buffers ();

  gVO.width = width;
  gVO.height = height;

  /* rate code */

  switch (rate_code) {
  case 1: /* 23.976 fps */
    gVO.video_step = 3913;
    break;
  case 2: /* 24 fps */
    gVO.video_step = 3750;
    break;
  case 3: /* 25 fps */
    gVO.video_step = 3600;
    break;
  case 4: /* 29.97 fps */
    gVO.video_step = 3003;
    break;
  case 5: /* 30 fps */
    gVO.video_step = 3000;
    break;
  case 6: /* 50 fps */
    gVO.video_step = 1800;
    break;
  case 7: /* 59.94 fps */
    gVO.video_step = 1525;
    break;
  case 8: /* 60 fps */
    gVO.video_step = 1509;
    break;
  default:
    fprintf (stderr, "invalid/unknown frame rate code : %d \n",
             rate_code);
    exit (1);
  }

  gVO.pts_per_half_frame = gVO.video_step/2;

  metronom_set_video_rate (gVO.video_step);
}

vo_image_buffer_t* vo_alloc_image_buffer() {
  vo_image_buffer_t *img;

  
  xprintf (VERBOSE|VIDEO, "video_out : alloc image buffer\n");
  img = remove_from_img_buf_queue (gVO.free_img_buf_queue);

  if (!img) {
    img = remove_from_img_buf_queue (gVO.display_img_buf_queue);
    /* xprintf (VERBOSE|VIDEO, "warning : no free img buffers left\n"); */
    printf ("warning : no free img buffers left\n");
    metronom_reset();
  }

  if (img) {
    img->bDisplayLock = 0;
    img->bDecoderLock = 1;
  }

  gVO.bInitialized = 1;

  return img;
}

void vo_free_image_buffer(vo_image_buffer_t *img) {
  img->bDecoderLock = 0;

  if (!img->bDisplayLock) 
    append_to_img_buf_queue (gVO.free_img_buf_queue, img);
}

void vo_queue_frame (vo_image_buffer_t* img) {

  int32_t  diff;
  uint32_t cur_vpts;
  uint32_t pic_vpts ;

  pic_vpts = metronom_got_video_frame (img->PTS);
  img->PTS = pic_vpts;
  gVO.num_frames_delivered++;

  xprintf (VERBOSE|VIDEO,"video_out: got image. vpts for picture is %d\n", pic_vpts);

  cur_vpts = gAudioOut->get_current_pts();

  diff = pic_vpts - cur_vpts;

  xprintf (VERBOSE|VIDEO,"video_out:: delivery diff : %d\n",diff);

  if (cur_vpts>0) {

    if (diff < gVO.video_step*3 ) {
      if (diff < gVO.video_step) {
	mpeg2_set_skip_frame (); 
	
	xprintf (VERBOSE|VIDEO,"      ======> skip frame\n");
	
	if (diff < 0) { 
	  xprintf (VERBOSE|VIDEO, "               ======> rejected\n");
	  
	  gVO.num_frames_discarded++;
	  return; 
	} 
	
      } else {
	mpeg2_set_skip_bframe (); 
	
	xprintf (VERBOSE|VIDEO,"      ======> skip bframe\n");
      }
    }
  } /* else: we are probably in precaching mode */

  if (!img->bFrameBad) {
    /*
     * put frame into FIFO-Buffer
     */

    xprintf (VERBOSE|VIDEO, "frame is ok => appending to display buffer\n");

    img->bDisplayLock = 1;
    
    append_to_img_buf_queue (gVO.display_img_buf_queue, img);

  } else
    gVO.num_frames_skipped++;

  /*
   * performance measurement
   */
  
  if (gVO.num_frames_delivered>199) {
    printf ("%d frames delivered, %d frames skipped, %d frames discarded\n", 
	    gVO.num_frames_delivered, gVO.num_frames_skipped, gVO.num_frames_discarded);

    gVO.num_frames_delivered = 0;
    gVO.num_frames_discarded = 0;
    gVO.num_frames_skipped   = 0;
  }
}

void vo_set_fullscreen (int bFullscreen) {
  gVideoDriver->set_fullscreen (bFullscreen);
}

int vo_is_fullscreen () {
  return gVideoDriver->is_fullscreen ();
}

int vo_is_buffer_full () {
  
  xprintf (VERBOSE|VIDEO, "video_out: is_buffer_full => %d\n", gVO.free_img_buf_queue->num_buffers);
  
  return gVO.bInitialized && (gVO.free_img_buf_queue->num_buffers <2);
}

void vo_video_out () {

  uint32_t           cur_pts;
  int                pts_absdiff, diff, absdiff, pts=0;
  vo_image_buffer_t *img;

  pts_absdiff = 1000000;

  cur_pts = gAudioOut->get_current_pts ();

  xprintf (VERBOSE|VIDEO, "video_out : vo_video_out called at audio pts %d\n", cur_pts);

  /*
   * print image queue
   */

  img = gVO.display_img_buf_queue->first;

  while (img) {
    pts = img->PTS;
    xprintf (VERBOSE|VIDEO, "video_out : display queue contains image with pts %d\n",pts);
    
    img = img->next;
  }


  img = gVO.display_img_buf_queue->first;

  if (!img)
    return;

  /*
   * throw away expired frames
   */
  
  do {
    pts = img->PTS;
    diff = cur_pts - pts;
    absdiff = abs(diff);

    xprintf (VERBOSE|VIDEO, "video_out : display queue contains image with pts %d\n",pts);

    if (diff >gVO.pts_per_half_frame) {
    
      /*
      xprintf (VERBOSE|VIDEO, "video_out : throwing away image with pts %d because "
	       "it's too old (diff : %d > %d).\n",pts,diff,
	       gVO.pts_per_half_frame);
	       */
      printf ("video_out : throwing away image with pts %d because "
	      "it's too old (diff : %d > %d).\n",pts,diff,
	      gVO.pts_per_half_frame);
      gVO.num_frames_discarded++;

      img = remove_from_img_buf_queue (gVO.display_img_buf_queue);
    
      if (!img->bDecoderLock) 
	append_to_img_buf_queue (gVO.free_img_buf_queue, img);

      img->bDisplayLock = 0;
    
      img = gVO.display_img_buf_queue->first;

      if (!img)
	return;
    }
  } while (diff >gVO.pts_per_half_frame); 

  /*
   * time to display frame 0 ?
   */

  if (diff<0)
    return;


  /*
   * remove frame from display queue and show it
   */

  xprintf (VERBOSE|VIDEO, "video_out : displaying image with pts = %d (diff=%d)\n", pts, diff);

  img = remove_from_img_buf_queue (gVO.display_img_buf_queue);
  profiler_start_count (5);
  gVideoDriver->display_frame (img);
  profiler_stop_count (5);
  
  /*
   * free image
   */

  if (!img->bDisplayLock)
    xprintf (VERBOSE|VIDEO, "video_out: ALERT! frame was not locked for display queue\n");
  img->bDisplayLock = 0;

  if (!img->bDecoderLock) {
    append_to_img_buf_queue (gVO.free_img_buf_queue, img);
  }    

}

void vo_handle_event (XEvent *event) {
  gVideoDriver->handle_event (event);
}

void vo_set_logo_mode (int bLogoMode) {
  gVideoDriver->set_logo_mode (bLogoMode);
}

void vo_reset (void) {

  vo_image_buffer_t *img;

  while ((img = remove_from_img_buf_queue (gVO.display_img_buf_queue))) {

    if (!img->bDisplayLock)
      xprintf (VERBOSE|VIDEO, "ALERT! frame was not locked in display queue\n");
    img->bDisplayLock = 0;

    if (!img->bDecoderLock) {

      append_to_img_buf_queue (gVO.free_img_buf_queue, img);
    }       
  }

  gVO.num_frames_delivered = 0;
  gVO.num_frames_skipped   = 0;
  gVO.num_frames_discarded = 0;
  gVO.bInitialized         = 0;   
}

void vo_display_cursor (int state) {

  gVideoDriver->hideshow_cursor(state);
}

void vo_set_aspect (int ratio) {
  gVideoDriver->set_aspect (ratio);
  reallocate_img_buffers ();
}

int vo_get_aspect () {
  return gVideoDriver->get_aspect ();
}

void vo_init (int no_xv) {

  gVideoDriver = NULL;

  gVO.width                 = 720;
  gVO.height                = 576;
  gVO.num_frames_delivered  = 0;
  gVO.num_frames_skipped    = 0;
  gVO.num_frames_discarded  = 0;
  gVO.free_img_buf_queue    = new_img_buf_queue ();
  gVO.display_img_buf_queue = new_img_buf_queue ();
  gVO.displayed_image       = NULL;
  gVO.bInitialized          = 0;   

#ifdef HAVE_XV
  if (!no_xv) {
    gVideoDriver = init_video_out_xv ();
    if (gVideoDriver) {
      printf ("Using X Window System video extension for video output.\n");

      vo_set_size (720,576,1,3); 

      return ;
    }
  }
#endif
  
  gVideoDriver = init_video_out_xshm ();
  if (gVideoDriver) {
    printf ("Using X Window System shared memory extension for video output.\n");

    vo_set_size (720,576,1,3); 

    return ;
  }
  

  /*
  vo = init_video_out_x11 ();
  if (vo) {
    printf ("Using plain X11 for video output.\n");
    return vo;
  }
  */

  /* vo_set_size (720,576,1,3); */

  fprintf (stderr, "\nAll video output plugins failed. Exit.\n\n");
  exit (1);
}
